from datetime import timedelta
from textwrap import dedent

from django.conf import settings
from django.utils import timezone
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import ListCreateAPIView, RetrieveDestroyAPIView
from rest_framework.response import Response

from arkindex.documents.models import Corpus, CorpusExport, CorpusExportState
from arkindex.documents.serializers.export import CorpusExportSerializer
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Role


@extend_schema(tags=["exports"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListExports",
        description=(
            "List all exports on a corpus.\n\n"
            "Guest access is required on private corpora."
        ),
    ),
    post=extend_schema(
        operation_id="StartExport",
        request=None,
        description=dedent(
            f"""
            Start a corpus export job.

            A user must wait for {settings.EXPORT_TTL_SECONDS} seconds after the last successful import
            before being able to generate a new export of the same corpus.

            Contributor access is required.
            """
        ),
    )
)
class CorpusExportAPIView(CorpusACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = CorpusExportSerializer
    queryset = CorpusExport.objects.none()

    def get_queryset(self):
        return CorpusExport \
            .objects \
            .filter(corpus=self.get_corpus(self.kwargs["pk"])) \
            .select_related("user") \
            .order_by("-created")

    def post(self, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs["pk"], role=Role.Contributor)

        if corpus.exports.filter(state__in=(CorpusExportState.Created, CorpusExportState.Running)).exists():
            raise ValidationError("An export is already running for this corpus.")

        available_exports = corpus.exports.filter(
            state=CorpusExportState.Done,
            created__gte=timezone.now() - timedelta(seconds=settings.EXPORT_TTL_SECONDS)
        )
        if available_exports.exists():
            raise ValidationError(f"An export has already been made for this corpus in the last {settings.EXPORT_TTL_SECONDS} seconds.")

        export = corpus.exports.create(user=self.request.user)
        export.start()
        return Response(CorpusExportSerializer(export).data, status=status.HTTP_201_CREATED)


@extend_schema(
    tags=["exports"],
)
@extend_schema_view(
    get=extend_schema(
        operation_id="DownloadExport",
        description=dedent(
            """
            Download a corpus export.

            Guest access is required on private corpora.
            """
        ),
        responses={302: serializers.Serializer},
    ),
    delete=extend_schema(
        operation_id="DestroyExport",
        description=dedent(
            """
            Delete a corpus export.

            Requires either an admin access to the corpus, or for the user to be
            the export's creator and have contributor access to the corpus.
            """
        )
    )
)
class ManageExport(RetrieveDestroyAPIView):
    queryset = CorpusExport.objects.none()
    permission_classes = (IsVerified, )
    serializer_class = CorpusExportSerializer

    def get_queryset(self):
        states = {CorpusExportState.Done}
        if self.request.method not in permissions.SAFE_METHODS:
            states.add(CorpusExportState.Running)
        return CorpusExport.objects.select_related("corpus").filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            state__in=states
        )

    def get(self, *args, **kwargs):
        return Response(status=status.HTTP_302_FOUND, headers={"Location": self.get_object().s3_url})

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method in permissions.SAFE_METHODS:
            return
        if not obj.is_deletable(request.user):
            raise PermissionDenied(detail="You do not have sufficient rights to delete this export.")
        # Allow deleting running exports if they have not been updated in longer than EXPORT_TTL_SECONDS (not actually still running)
        if obj.state == CorpusExportState.Running and obj.updated + timedelta(seconds=settings.EXPORT_TTL_SECONDS) > timezone.now():
            raise ValidationError("You cannot delete an export that is still running.")
