from collections import defaultdict

from django.db.models import Exists, OuterRef
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Task
from arkindex.process.models import Process, ProcessDataset, ProcessMode
from arkindex.project.mixins import ProcessACLMixin
from arkindex.training.models import Dataset
from arkindex.users.models import Role


class ProcessDatasetSerializer(ProcessACLMixin, serializers.ModelSerializer):
    process = serializers.PrimaryKeyRelatedField(
        queryset=(
            Process
            .objects
            # Avoid a stale read when adding a dataset right after creating a process
            .using("default")
            # Required for ACL checks
            .select_related("corpus")
            # Required to check for a process that have already started
            .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk"))))
        ),
        style={"base_template": "input.html"},
    )
    dataset = serializers.PrimaryKeyRelatedField(
        queryset=Dataset.objects.none(),
        style={"base_template": "input.html"},
    )

    class Meta:
        model = ProcessDataset
        fields = ("dataset", "process", "id", )
        read_only_fields = ("process", "id", )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Do not raise Error in order to create OpenAPI schema
            return

        request_method = self.context["request"].method
        # Required for the ProcessACLMixin and readable corpora
        self._user = self.context["request"].user

        if request_method == "DELETE":
            # Allow deleting ProcessDatasets even if the user looses access to the corpus
            self.fields["dataset"].queryset = Dataset.objects.all()
        else:
            self.fields["dataset"].queryset = Dataset.objects.filter(corpus__in=Corpus.objects.readable(self._user))

    def validate_process(self, process):
        access = self.process_access_level(process)
        if not access or not (access >= Role.Admin.value):
            raise PermissionDenied(detail="You do not have admin access to this process.")
        return process

    def validate(self, data):
        process, dataset = data["process"], data["dataset"]
        errors = defaultdict(list)

        if process.mode != ProcessMode.Dataset:
            errors["process"].append('Datasets can only be added to or removed from processes of mode "dataset".')

        if process.has_tasks:
            errors["process"].append("Datasets cannot be updated on processes that have already started.")

        if self.context["request"].method == "POST" and process.datasets.filter(id=dataset.id).exists():
            errors["dataset"].append("This dataset is already selected in this process.")

        if errors:
            raise ValidationError(errors)

        return data
