"""
Helper methods to trigger tasks in asynchronous workers
"""
from typing import Literal, Optional, Union
from uuid import UUID

from django.db.models import Prefetch, prefetch_related_objects

from arkindex.documents import export
from arkindex.documents import tasks as documents_tasks
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import Corpus, CorpusExport, Element
from arkindex.ponos import tasks as ponos_tasks
from arkindex.ponos.models import State, Task
from arkindex.process import tasks as process_tasks
from arkindex.process.models import (
    Process,
    ProcessMode,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerRun,
    WorkerVersion,
)
from arkindex.training.models import ModelVersion


def corpus_delete(corpus: Union[Corpus, UUID, str], user_id: Optional[int] = None) -> None:
    """
    Delete a whole corpus without killing a server by removing all related
    models first and lowering the amount of foreign keys Django has to handle.
    """
    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
        description = f"Deletion of corpus {corpus.name}"
    else:
        corpus_id = str(corpus)
        description = f"Deletion of corpus {corpus_id}"

    documents_tasks.corpus_delete.delay(corpus_id=corpus_id, description=description, user_id=user_id)


def element_trash(queryset: ElementQuerySet,
                  delete_children: bool = True,
                  user_id: Optional[int] = None) -> None:
    """
    Run ElementQuerySet.trash to delete a batch of elements.
    """
    assert isinstance(queryset, ElementQuerySet), "Only Element querysets can be trashed"
    documents_tasks.element_trash.delay(
        queryset=queryset,
        delete_children=delete_children,
        user_id=user_id,
        description="Element deletion",
    )


def worker_results_delete(corpus_id: UUID,
                          worker_run: Optional[WorkerRun] = None,
                          version: Optional[WorkerVersion] = None,
                          element_id: Optional[UUID] = None,
                          model_version: Optional[ModelVersion] = None,
                          configuration: Optional[WorkerConfiguration | Literal[False]] = None,
                          user_id: Optional[int] = None) -> None:
    """
    Delete all Worker Results produced by a specific WorkerRun or WorkerVersion
    (+ ModelVersion and Configuration), or any worker, on a whole corpus or under
    a specified parent element (parent element included).
    """
    description = "Deletion of worker results"
    if worker_run is not None:
        description += f" produced by {worker_run.summary}"
    else:
        if version is not None:
            description += f" produced by {version}"
        if model_version is not None:
            description += f" based on model version {model_version}"
        if configuration is False:
            description += " with no configuration"
        elif configuration is not None:
            description += f" configured with {configuration}"

    documents_tasks.worker_results_delete.delay(
        corpus_id=corpus_id,
        element_id=element_id,
        worker_run_id=worker_run.id if worker_run else None,
        version_id=version.id if version else None,
        model_version_id=model_version.id if model_version else None,
        configuration_id=configuration.id if configuration else configuration,
        user_id=user_id,
        description=description,
    )


def selection_worker_results_delete(corpus: Corpus,
                                    user_id: int = None,
                                    worker_run: Optional[WorkerRun] = None,
                                    version: Optional[WorkerVersion] = None,
                                    model_version: Optional[ModelVersion] = None,
                                    configuration: Optional[WorkerConfiguration | Literal[False]] = None) -> None:
    """
    Delete all Worker Results by a specific WorkerRun or WorkerVersion
    (+ ModelVersion and Configuration), or any worker, on all elements
    selected by a user in a corpus, including their children elements.
    """
    description = f"Deletion of worker results on selected elements in {corpus.name}"
    if worker_run is not None:
        description += f" produced by {worker_run.summary}"
    else:
        if version is not None:
            description += f" produced by {version}"
        if model_version is not None:
            description += f" based on model version {model_version}"
        if configuration is False:
            description += " with no configuration"
        elif configuration is not None:
            description += f" configured with {configuration}"

    documents_tasks.selection_worker_results_delete.delay(
        corpus_id=corpus.id,
        worker_run_id=worker_run.id if worker_run else None,
        version_id=version.id if version else None,
        model_version_id=model_version.id if model_version else None,
        configuration_id=configuration.id if configuration else configuration,
        user_id=user_id,
        description=description,
    )


def move_element(source: Element, destination: Element, user_id: Optional[int] = None) -> None:
    """
    Move a source Element (and all of its children) to a destination Element.
    Delete all source Element previous parents.
    """
    documents_tasks.move_element.delay(
        source=source,
        destination=destination,
        user_id=user_id,
        description=f"Moving element {source.name} to element {destination.name}"
    )


def move_selection(corpus_id: UUID, destination: Element, user_id: int) -> None:
    """
    Move selected elements (and all of their children) to a destination Element.
    Delete all those elements' previous parents.
    """
    documents_tasks.move_selection.delay(
        corpus_id=corpus_id,
        destination=destination,
        user_id=user_id,
        description=f"Moving selected elements from corpus {corpus_id} to element {destination.name}"
    )


def add_parent_selection(corpus_id: UUID, parent: Element, user_id: int) -> None:
    """
    Add a parent element to the selected elements without removing the other existing parents.
    """
    documents_tasks.add_parent_selection.delay(
        corpus_id=corpus_id,
        parent=parent,
        user_id=user_id,
        description=f"Adding {parent.name} as parent of selected elements from corpus {corpus_id}",
    )


def reindex_corpus(corpus_id: UUID, user_id: int, drop: bool = True, job_id: Optional[str] = None) -> None:
    """
    Index elements of a given project
    """
    documents_tasks.reindex_corpus.delay(
        corpus_id=corpus_id,
        drop=drop,
        user_id=user_id,
        description=f"Reindexing project {corpus_id}",
        job_id=job_id,
    )


def initialize_activity(process: Process):
    """
    Initialize activity on every process elements for worker versions that are part of its workflow
    """
    process_tasks.initialize_activity.delay(process)


def export_corpus(corpus_export: CorpusExport) -> None:
    """
    Export a corpus to a SQLite database
    """
    export.export_corpus.delay(
        corpus_export=corpus_export,
        user_id=corpus_export.user_id,
        description=f"Export of corpus {corpus_export.corpus.name}"
    )


def process_delete(process: Process, delete_activity: Optional[WorkerActivityState] = None):
    """
    Delete a process with its related worker activities
    """
    process_tasks.process_delete.delay(process, delete_activity=delete_activity)


def notify_process_completion(process: Process):
    """
    Notify the creator by email when a process reach a specific state.
    Processes with Repository mode are skipped.
    """
    prefetch_related_objects([process], Prefetch(
        "tasks",
        # Avoid a stale read, since we just updated a task's state
        Task.objects.using("default")
        # Pick only the tasks of the last run, computed in a subquery
        .filter(run=process.tasks.order_by("-run").values("run")[:1])
        # Sort the tasks to get a sorted output in the email
        .order_by("depth", "slug")
    ))
    state = process.state

    state_msg = {
        State.Completed: "successfully",
        State.Failed: "with failures",
        State.Error: "with errors",
        State.Stopped: "because it was stopped",
    }
    if process.mode != ProcessMode.Repository and state in state_msg.keys():
        process_name = process.name or str(process.id)
        ponos_tasks.notify_process_completion.delay(
            process=process,
            subject=f"Your process {process_name} finished {state_msg[state]}",
        )
